#include "framework.h"
#include "Scripting/Internal/TEN/Input/InputHandler.h"

#include "Renderer/RendererEnums.h"
#include "Scripting/Internal/LuaHandler.h"
#include "Scripting/Internal/ReservedScriptNames.h"
#include "Scripting/Internal/ScriptUtil.h"
#include "Scripting/Internal/TEN/Input/ActionIDs.h"
#include "Scripting/Internal/TEN/Types/Vec2/Vec2.h"
#include "Specific/Input/Input.h"

using namespace TEN::Input;

/// Functions for input management.
// @tentable Input
// @pragma nostrip

namespace TEN::Scripting::Input
{
	/// Vibrate the game controller if the function is available and the setting is on.
	// @function Vibrate
	// @tparam float strength Vibration strength. 
	// @tparam[opt=0.3] float time Vibration time in seconds.
	static void Vibrate(float strength, sol::optional<float> time)
	{
		Rumble(strength, time.value_or(0.3f), RumbleMode::Both);
	}

	static bool CheckInput(int actionID)
	{
		if (actionID > (int)InputActionID::Count)
		{
			ScriptAssertF(false, "Input action {} does not exist.", actionID);
			return false;
		}

		return true;
	}

	/// Check if an action key is being held.
	// @function KeyIsHeld
	// @tparam Input.ActionID action Action ID to check.
	static bool KeyIsHeld(int actionID)
	{
		if (!CheckInput(actionID))
			return false;

		if (IsHeld((InputActionID)actionID))
			return true;

		return false;
	}

	/// Check if an action key is being hit or clicked.
	// @function KeyIsHit
	// @tparam Input.ActionID action Action ID to check.
	static bool KeyIsHit(int actionID)
	{
		if (!CheckInput(actionID))
			return false;

		if (IsClicked((InputActionID)actionID))
			return true;

		return false;
	}

	/// Simulate an action key push.
	// @function KeyPush
	// @tparam Input.ActionID action Action ID to push.
	static void KeyPush(int actionID)
	{
		if (!CheckInput(actionID))
			return;

		ActionQueueMap[(InputActionID)actionID] = ActionQueueState::Update;
	}

	/// Clear an action key.
	// @function KeyClear
	// @tparam Input.ActionID action Action ID to clear.
	static void KeyClear(int actionID)
	{
		if (!CheckInput(actionID))
			return;

		ActionQueueMap[(InputActionID)actionID] = ActionQueueState::Clear;
	}

	/// Clear all action keys.
	// @function KeyClearAll
	static void KeyClearAll()
	{
		for (auto& [actionID, queue] : ActionQueueMap)
			queue = ActionQueueState::Clear;
	}

	/// Get the display position of the cursor in percent.
	// @function GetMouseDisplayPosition
	// @treturn Vec2 Cursor display position in percent.
	static Vec2 GetMouseDisplayPosition()
	{
		// NOTE: Conversion from internal 800x600 to more intuitive 100x100 display space resolution is required.
		// In a future refactor, everything will use 100x100 natively. -- Sezz 2023.10.20

		auto cursorPos = GetMouse2DPosition();
		cursorPos = Vector2(cursorPos.x / DISPLAY_SPACE_RES.x, cursorPos.y / DISPLAY_SPACE_RES.y) * 100;
		return Vec2(cursorPos);
	}

	void Register(sol::state* state, sol::table& parent)
	{
		auto table = sol::table(state->lua_state(), sol::create);

		parent.set(ScriptReserved_Input, table);
		table.set_function(ScriptReserved_Vibrate, &Vibrate);
		table.set_function(ScriptReserved_KeyIsHeld, &KeyIsHeld);
		table.set_function(ScriptReserved_KeyIsHit, &KeyIsHit);
		table.set_function(ScriptReserved_KeyPush, &KeyPush);
		table.set_function(ScriptReserved_KeyClear, &KeyClear);
		table.set_function(ScriptReserved_KeyClearAll, &KeyClearAll);

		table.set_function(ScriptReserved_GetMouseDisplayPosition, &GetMouseDisplayPosition);
		table.set_function(ScriptReserved_GetCursorDisplayPosition, &GetMouseDisplayPosition);

		auto handler = LuaHandler(state);
		handler.MakeReadOnlyTable(table, ScriptReserved_ActionID, ACTION_IDS);
	}
}
